"""Root conftest.py file for pytest configuration."""

# pylint: disable=redefined-outer-name

import os
from unittest import mock
import uuid

from pytest import fixture

from drift_monitor import DriftMonitor


@fixture(scope="session")
def endpoint():
    """Return the server URL."""
    return os.environ["DRIFT_MONITOR_URL"]


@fixture(scope="session")
def token(request):
    """Return the server token."""
    token = request.param if hasattr(request, "param") else "1234"
    return token


@fixture(scope="function")
def request_mock():
    """Patch requests module with MagicMocks."""
    with mock.patch("drift_monitor.utils.requests") as requests:
        yield requests


@fixture(scope="function")
def monitor(token):
    """Return a DriftMonitor instance."""
    return DriftMonitor("model_1", token)


@fixture(scope="function", autouse=True)
def post_response(request, request_mock, monitor):
    """Return a POST response and patch it in request_mock."""
    if hasattr(request, "param"):
        json = request.param
    else:
        json = {
            "id": f"{uuid.uuid4()}",
            "datetime": "2021-01-01T00:00:00Z",
            "model_id": monitor.model_id,
            "status": "Running",
        }
    request_mock.post.return_value = mock.MagicMock(json=lambda: json)
    return json


@fixture(scope="function")
def with_context(monitor):
    """Opens a context for the monitor."""
    with monitor:
        yield


@fixture(scope="function")
def after_context(monitor):
    """Closes the context for the monitor."""
    with monitor:
        monitor.concept(True, {"threshold": 0.5})
        monitor.data(True, {"threshold": 0.5})


@fixture(scope="function")
def error_context(monitor):
    """Raise an error in the context for the monitor."""
    try:
        with monitor:
            raise ValueError("An error occurred.")
    except ValueError:
        pass


@fixture(scope="function")
def with_concept_drift(monitor):
    """Add concept drift to the monitor."""
    monitor.concept(True, {"threshold": 0.5})


@fixture(scope="function")
def with_data_drift(monitor):
    """Add data drift to the monitor."""
    monitor.data(True, {"threshold": 0.5})



"""Configuration module for drift_monitor detection client."""

from libmytoken import MytokenServer
from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    """Settings class for drift_monitor detection client."""

    model_config = SettingsConfigDict(
        case_sensitive=False,
    )

    DRIFT_MONITOR_DOMAIN: str
    DRIFT_MONITOR_TIMEOUT: int = 10

    DRIFT_MONITOR_MYTOKEN: str
    MYTOKEN_SERVER: str = "https://mytoken.data.kit.edu"

    TESTING: bool = False


# Initialize the settings object
settings = Settings()
mytoken_server = MytokenServer(settings.MYTOKEN_SERVER)



"""Test example module."""

import pytest


def test_concept_context(monitor):
    """Test the method concept raises out of context error."""
    with pytest.raises(RuntimeError) as excinfo:
        monitor.concept(True, {"threshold": 0.5})
    assert str(excinfo.value) == "Drift monitor context not started."


def test_data_context(monitor):
    """Test the method concept raises out of context error."""
    with pytest.raises(RuntimeError) as excinfo:
        monitor.data(True, {"threshold": 0.5})
    assert str(excinfo.value) == "Drift monitor context not started."



"""Utility functions for drift monitor."""

import requests

from drift_monitor.config import settings


def create_drift(model, token):
    """Create a drift run on the drift monitor server."""
    response = requests.post(
        url=f"https://{settings.DRIFT_MONITOR_DOMAIN}/api/drift",
        headers={"Authorization": f"Bearer {token}"},
        json={"model": model, "job_status": "Running"},
        timeout=settings.DRIFT_MONITOR_TIMEOUT,
        verify=not settings.TESTING,
    )
    response.raise_for_status()
    return response.json()


def complete_drift(drift, token):
    """Complete a drift run on the drift monitor server."""
    _drift = {k: v for k, v in drift.items() if k != "id" and k != "datetime"}
    response = requests.put(
        url=f"https://{settings.DRIFT_MONITOR_DOMAIN}/api/drift/{drift['id']}",
        headers={"Authorization": f"Bearer {token}"},
        json={**_drift, "job_status": "Completed"},
        timeout=settings.DRIFT_MONITOR_TIMEOUT,
        verify=not settings.TESTING,
    )
    response.raise_for_status()


def fail_drift(drift, token):
    """Fail a drift run on the drift monitor server."""
    _drift = {k: v for k, v in drift.items() if k != "id" and k != "datetime"}
    response = requests.put(
        url=f"https://{settings.DRIFT_MONITOR_DOMAIN}/api/drift/{drift['id']}",
        headers={"Authorization": f"Bearer {token}"},
        json={**_drift, "job_status": "Failed"},
        timeout=settings.DRIFT_MONITOR_TIMEOUT,
        verify=not settings.TESTING,
    )
    response.raise_for_status()


def register(token):
    """Registers the token user in the application database."""
    response = requests.post(
        url=f"https://{settings.DRIFT_MONITOR_DOMAIN}/api/user",
        headers={"Authorization": f"Bearer {token}"},
        timeout=settings.DRIFT_MONITOR_TIMEOUT,
        verify=not settings.TESTING,
    )
    response.raise_for_status()



"""Drift Monitor Client package.
This package contains the client code for the drift monitor service.
"""

import requests
from drift_monitor import utils
from drift_monitor.config import mytoken_server, settings


class DriftMonitor:
    """Drift Monitor context.
    This class is a context manager for the drift monitor service. It is used
    as an abstraction for the user to interact with the drift monitor service.

    When the context is entered, the drift monitor sends a POST request to the
    server to create a drift run. When the context is exited, the drift monitor
    sends a PUT request to the server to complete the drift run.

    Args:
        model_id (str): The model ID to monitor.
        token (str): The token to authenticate with the server.
            Defaults to the token in the settings.

    Example:
        >>> with DriftMonitor("model_1", token="123") as monitor:
        ...    detected, detection_parameters = concept_detector()
        ...    monitor.concept(detected, detection_parameters)
        ...    detected, detection_parameters = data_detector()
        ...    monitor.data(detected, detection_parameters)
    """

    def __init__(self, model_id, token=settings.DRIFT_MONITOR_MYTOKEN):
        self.model_id = model_id
        self.at = mytoken_server.AccessToken.get(token)
        self.drift = None

    def concept(self, detected, parameters):
        """Prepare concept drift detection results to the server.

        Args:
            detected (bool): Whether concept drift was detected.
            detection_parameters (dict): The parameters used for detection.

        Raises:
            RuntimeError: If the drift monitor context is not started.
        """
        if self.drift is None:
            raise RuntimeError("Drift monitor context not started.")
        concept_drift = {"drift": detected, "parameters": parameters}
        self.drift["concept_drift"] = concept_drift

    def data(self, detected, parameters):
        """Prepare data drift detection results to the server.

        Args:
            detected (bool): Whether data drift was detected.
            detection_parameters (dict): The parameters used for detection.

        Raises:
            RuntimeError: If the drift monitor context is not started.
        """
        if self.drift is None:
            raise RuntimeError("Drift monitor context not started.")
        data_drift = {"drift": detected, "parameters": parameters}
        self.drift["data_drift"] = data_drift

    def __enter__(self):
        self.drift = utils.create_drift(self.model_id, self.at)
        return self

    def __exit__(self, exc_type, _exc_value, _traceback):
        if exc_type:
            utils.fail_drift(self.drift, self.at)
        else:
            utils.complete_drift(self.drift, self.at)


def register(accept_terms=False, token=settings.DRIFT_MONITOR_MYTOKEN):
    """Registers the token user in the application database.
    By using this function, you accept that the user derived from the token
    will be registered in the application database and agree to the terms of
    service.
    """
    if not accept_terms:
        raise ValueError("You must accept the terms of service.")
    try:
        utils.register(mytoken_server.AccessToken.get(token))
    except requests.HTTPError as error:
        if error.response.status_code == 409:
            return  # User already exists
        raise error



"""Test example module."""

import pytest


@pytest.mark.usefixtures("with_context")
def test_running_drift(request_mock, endpoint, token):
    """Test the POST request to create a drift run was sent to server."""
    assert request_mock.post.call_count == 1
    url = f"{endpoint}/drift"
    assert request_mock.post.call_args[1]["url"] == url
    assert request_mock.post.call_args[1]["headers"] == {
        "Authorization": f"Bearer {token}"
    }
    assert request_mock.post.call_args[1]["json"] == {
        "model_id": "model_1",
        "status": "Running",
    }


@pytest.mark.usefixtures("after_context")
def test_completed_drift(request_mock, endpoint, token, monitor):
    """Test the PUT request to complete a drift run was sent to server."""
    assert request_mock.put.call_count == 1
    url = f"{endpoint}/drift/{monitor.drift['id']}"
    assert request_mock.put.call_args[1]["url"] == url
    assert request_mock.put.call_args[1]["headers"] == {
        "Authorization": f"Bearer {token}"
    }
    assert request_mock.put.call_args[1]["json"] == {
        **monitor.drift,
        "status": "Completed",
        "concept_drift": {"drift": True, "parameters": {"threshold": 0.5}},
        "data_drift": {"drift": True, "parameters": {"threshold": 0.5}},
    }


@pytest.mark.usefixtures("error_context")
def test_failed_drift(request_mock, endpoint, token, monitor):
    """Test the PUT request to fail a drift run was sent to server."""
    assert request_mock.put.call_count == 1
    url = f"{endpoint}/drift/{monitor.drift['id']}"
    assert request_mock.put.call_args[1]["url"] == url
    assert request_mock.put.call_args[1]["headers"] == {
        "Authorization": f"Bearer {token}"
    }
    assert request_mock.put.call_args[1]["json"] == {
        **monitor.drift,
        "status": "Failed",
    }
